/** 
*  \file SimCal3Test.cc
*  \brief Unit test for calibration for simulation data
*    
*  \author G. Martin <graemem@ssec.wisc.edu>
*
*  \version $Id: SimCal3Test.cc,v 1.6.2.14 2006/01/13 18:29:10 rayg Exp $
*
*  \par Copyright:
*  \verbatim
*
*  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
*  Space Science and Engineering Center
*  University of Wisconsin - Madison, USA
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* FUTURE: consider parallelizing file I/O. Possibly group readers and 
* writers separately within a Binder Group and specify no sharing of 
* resources among different binders to prevent race conditions. Binders
* can themselves be multi-threaded if desired.
*
*  \endverbatim
*/


/*****************************************************************
 QUICK OUTLINE FOR SIMCAL3

STUDY sub-pipeline: H, A, Z scenes

    IfgToSpectrum -> PiecewiseLinearCalibrationStudy

    This loads the calibration context cache with context instances which can be used to calibrate earth data.

APPLY sub-pipeline: E scenes

    (IfgToSpectrum; ThreeBodyPolynomialCalibrationSynthesis) -> ThreeBodyRadiometricCalibration -> Resampler

    This uses the data available in the context cache in order to synthesize (interpolate in this case) the blackbody radiance data 
    which feeds the 3-body calibration equation.

******************************************************************/



#include <cassert>
#include <list>
#include <algorithm>
#include <memory> // auto_ptr
#include <string>
#include <vector>
#include <iostream> 
//typedef boost::shared_ptr<FileToMemBinder> FileToMemBinder_P;
// std::cerr
#include <fstream>
#include <math.h>

// NETCDF includes
#include "netcdf.hh"
#include "netcdfcpp.h"

#include "Physics.h"  //wnums_t
#include "Planck.hxx"

#include "IfgToSpectrum.hxx"
    // provides IfgToSpectrum<>
#include "PiecewiseLinearCalibrationContext.hxx"  
    // provides PiecewiseLinearCalibrationStudy<>, PolynomialCalibrationContext, ThreeBodyPolynomialCalibrationSynthesis<>
#include "RadiometricCalibration.hxx"
    // provides ThreeBodyRadiometricCalibration<>
#include "Resample.hxx" 
    // provides SpectralResampling<>
#include "BindCdf.hxx"  // bring in source data and reference settings

using namespace std;
using namespace gips;

const double HBB_TEMP = 300.0;
const double ABB_TEMP = 265.0;
const double ZBB_TEMP = 2.76;

const unsigned IFG_SIZE=2048;  
const unsigned SPC_SIZE=IFG_SIZE/2;
const unsigned NUM_RECS = 16384;
typedef vector< complex <double> > CplxSpectrum;
typedef vector<double> RealSpectrum;

typedef IfgToSpectrum< RealSpectrum, CplxSpectrum > Ifg2Spc;
typedef PolynomialCalibrationContext::BasicCacheType ContextCache;
typedef PiecewiseLinearCalibrationStudy< CplxSpectrum, ContextCache > CalStudy;
typedef ThreeBodyPolynomialCalibrationSynthesis< ContextCache, CplxSpectrum > CalSynth;
// three-body calibration includes as reference an emissivity spectrum
typedef ThreeBodyRadiometricCalibration< CplxSpectrum, RealSpectrum, physics::wnums_t, vector< double > > CalThreeBody;
typedef SpectralResampling<RealSpectrum,RealSpectrum> Resampler;

typedef NcVar *NcVar_p; // NOT! boost::shared_ptr<NcVar> NcVar_p; since NcFile retains ownership!

typedef boost::shared_ptr<FileToMemBinder> FileToMemBinder_P;

// Data is loaded into the test bench, which is connected to the assortment of components needed
// to achieve the processing goal. Each component has limited visibility of the work area, with
// sub-structures satisfying their connection requirements. This reduces copying and dynamic allocation
// substantially, and permits "syntax-compatible" data containers to be used according to the situation;
// for instance, concurrency-controlled containers could be considered in the case that the
// components (e.g. the input service) are operating in multiple threads.



struct TestBench
{   
    // context cache holds the data needed to synthesize time-matched blackbody "observations"
    ContextCache contexts;
    
    // ifg2spectrum input and output for all four scene types
    double theta; // theta angle off-axis, (milliradians)
    RealSpectrum obs;               // connectors
    CplxSpectrum obsSpectrum;   
    Ifg2Spc::settings_t obsI2sSettings;   // settings
    Ifg2Spc::Ports obsI2sPorts;           // ports

    CalStudy::settings_t studySettings;
    CalStudy::ancillary_t studyAncil;
    CalStudy::Ports studyPorts;

    // these buffers are used as the input to radiometric calibration and the output of the synthesizer    
    CplxSpectrum abbSynth;
    CplxSpectrum zbbSynth;   
    CplxSpectrum hbbSynth;
    
    CalSynth::settings_t synthSettings;
    CalSynth::ancillary_t synthAncil;
    CalSynth::Ports synthPorts;

    // Three body cal variables
    CalThreeBody::settings_t calSettings;  //settings      
    CalThreeBody::reference_t ref;
    CalThreeBody::ancillary_t ancil;    
    RealSpectrum cal, diag; 
    CalThreeBody::Ports calPorts;      // ports
    
    // Resampling variables
    Resampler::settings_t resampSettings;
    Resampler::ancillary_t resampAncil;
    RealSpectrum resampled;
    Resampler::Ports resampPorts;
    
    static physics::wnums_t wavenumbersLW( ) // from ROK, wavenumbers for spectral form; move to CDF reference file
    {
        const unsigned IfgSize = IFG_SIZE;
        const double vlaser = 9398.0; // effective laser wavenumber
        const double r_lw = 8; // reduction factor
        const double n_lw = IfgSize;
        const unsigned SpcSize = IfgSize/2;
        
        double x_lw = (n_lw/2.0) * r_lw/vlaser;
        double dv = 1.0 / (2.0 * x_lw);
        double v0_lw = (n_lw/2 + 1) * dv;
        physics::wnums_t ret; // FUTURE: make a friendlier constructor
        ret.baseWnum = v0_lw;
        ret.deltaWnum = dv;
        ret.wnumCount = SpcSize;
        return ret;
    }
    
    /// Convert theta value to a wavenumber ratio usable by Resampler
    static double wnumRatioFromTheta( double theta )
    {
        // FIXME: not yet implemented
        return 0.999;
    }

    // ctor
    TestBench():obs(IFG_SIZE), obsSpectrum(SPC_SIZE), obsI2sPorts(&obs, &obsSpectrum ),
                studyPorts( &obsSpectrum, &studyAncil, &contexts ),
                abbSynth(SPC_SIZE), 
                zbbSynth(SPC_SIZE), 
                hbbSynth(SPC_SIZE), 
                synthPorts( &synthAncil, &contexts, &hbbSynth, &abbSynth, &zbbSynth ),
                cal(SPC_SIZE), diag(SPC_SIZE), 
                calPorts( &ref, &ancil, &obsSpectrum, &hbbSynth, &abbSynth, &zbbSynth, &cal, &diag ),
                resampled(SPC_SIZE),
                resampPorts( &resampAncil, &cal, &resampled )
        {
            obsI2sSettings.interferogramSize = IFG_SIZE;   
            obsI2sSettings.chopRawSpectrumFrom = IFG_SIZE/2;
            obsI2sSettings.chopRawSpectrumTo = IFG_SIZE;
            obsI2sSettings.wrapRawSpectrumLastPoint = true; // FIXME: assess whether this is correct or not!! false causes 0.0 to be inserted, which causes NaNs to be generated in calibration, which blows up into NaNs throughout when Resampling occurs.. -RKG
            
            studySettings.spectrumSize = SPC_SIZE;
            synthSettings.spectrumSize = SPC_SIZE;
            calSettings.spectrumSize = SPC_SIZE;
            
            // FUTURE: move this to a reference data file (probably CDF) rather than generating it hard-coded
            ref.wavenumbers = wavenumbersLW( );
//            cout << "wavenumbers: base" << ref.wavenumbers.baseWnum 
//                    << " interval " << ref.wavenumbers.deltaWnum << endl;

            // FUTURE: load bbEmissivity from a bound CDF file properly interpolated to the wavenumbers
            // currently just fills with 1.0 emissivities and transmission
            ref.bbEmissivity.clear();
            ref.bbEmissivity.resize( SPC_SIZE, 1.0 );
            ref.telescopeTau = 1.0;
            ref.calmirrorTau = 1.0;
            
            // FUTURE: blackbody files should include temperatures over time
            // this is a reasonable approximation for simulation since BBs are controlled to tens of millikelvin
            ancil.abbTemp = ABB_TEMP;
            ancil.hbbTemp = HBB_TEMP;
            ancil.zbbTemp = ZBB_TEMP;
            ancil.reflectedTemp = 200.0;
            
            resampSettings.spectrumSize = SPC_SIZE;
            resampSettings.upsampledSpectrumSize = 32769; // FIXME: hard coded
        }    
};

// Not fully tested.
// extract the part of the string after the last '/'. If there
// there is no '/', return the entire string. If string ends
// in '/' return string with length 0;
string extractFilenameUnix(const string& str)
{
    unsigned slashPos = str.rfind('/');
    if(slashPos == (str.npos))    // no slash
        return string(str); 
    else if(slashPos == (str.length()-1))  // ends with slash
        return string(""); 
    else                          // string has a slash
        return string(str.substr(slashPos, str.length()-1));
}
 
struct FileEntry
{
    // used by both study and synth
    char scenetype; // H/A/Z/E
    double time; // must be linear/consistent/accurate measure of time
    
    // used by 3-body cal
    double hbbTemp;
    double abbTemp;
    double zbbTemp;
    double reflectedTemp;
    
    char scandir; // always same, arbitrary for sim data
    
    // used by control logic
    std::string filename;
    std::auto_ptr<NcFile> nc;
    
    FileEntry( const std::string &filename_, char scenetype_ = '\0' ):
        scenetype(scenetype_),
        filename(filename_),
        nc( new NcFile( filename_.c_str(), NcFile::ReadOnly ) )
    {
        // FUTURE: read these from the file or from a demo sequencing file..
        hbbTemp = HBB_TEMP;
        abbTemp = ABB_TEMP;
        zbbTemp = ZBB_TEMP;
        reflectedTemp = 200.0;
        scandir = 'F';
                
        // fetch time from the file, assuming one record per file
        // FUTURE: don't assume one record per file!
        { 
            NcError ncerr( NcError::verbose_nonfatal );

            NcVar *epoch = nc->get_var( "Epoch_time" );
            if (epoch!=NULL)
            {
                long count[] = { 1 };
                epoch->get( &time, &count[0] );
                std::cout << "time for " << filename_ << ": " << time << std::endl;
            }
            else
            {
                std::cerr << "WARNING: no Epoch_time in " << filename_ << ", setting to 0.0\n";
                time = 0.0;
            }
        }
                
        // FUTURE: if (!scenetype_) read-scenetype-from-file
    }
};

typedef std::list< boost::shared_ptr<FileEntry> > FileEntrySeq; // file entry sequence
// can't use std::auto_ptr in collections, not copy constructable


struct OutputPack
{
    std::auto_ptr<NcFile> fpCal, fpUncal, fpResampled;

    // binding of memory buffers to output file
    GiftsPixelRecordMapping mapping;    
    FileToMemBinderGroup mem2disk;

    OutputPack( const std::string &inputfilename, TestBench &myBench )
    {
        // create output filename, as awkwardly as possible
        // FUTURE: use std::ostringstream, you know you want to
        string calFileName("test/");
        calFileName.append(extractFilenameUnix(inputfilename).c_str());
        calFileName.append(".cal.nc");
        string uncalFileName("test/");
        uncalFileName.append(extractFilenameUnix(inputfilename).c_str());
        uncalFileName.append(".uncal.nc");
        string resampledFileName("test/");
        resampledFileName.append(extractFilenameUnix(inputfilename).c_str());
        resampledFileName.append(".resampled.nc");
        
        // create blank output file from template
        std::string ncgenCommand( "ncgen -b test/template.cdl -o ");
        int rc = system( (ncgenCommand + calFileName).c_str() );
        assert( 0 == rc );
        rc = system( (ncgenCommand + uncalFileName).c_str() );
        assert( 0 == rc );
        rc = system( (ncgenCommand + resampledFileName).c_str() );
        assert( 0 == rc );
    
        // create output files
        fpCal = std::auto_ptr<NcFile>( new NcFile( calFileName.c_str(), NcFile::Write ) ); 
        fpUncal = std::auto_ptr<NcFile>( new NcFile( uncalFileName.c_str(), NcFile::Write ) ); 
        fpResampled = std::auto_ptr<NcFile>( new NcFile( resampledFileName.c_str(), NcFile::Write ) ); 
    
        // add generated_by global attribute
        string tagName("$Id: SimCal3Test.cc,v 1.6.2.14 2006/01/13 18:29:10 rayg Exp $");
        fpCal->add_att("generated_by", tagName.c_str());
        fpUncal->add_att("generated_by", tagName.c_str());
        fpResampled->add_att("generated_by", tagName.c_str());
    
        long shape[] = { 1, 1, SPC_SIZE };
    
        mem2disk.push_back( newCdfCplxVectorWriter( &myBench.obsSpectrum, NcVar_p(fpUncal->get_var("signalReal_lw")), NcVar_p(fpUncal->get_var("signalImag_lw")), shape, shape+3, &mapping ));
        mem2disk.push_back( newCdfVectorWriter( &myBench.cal, NcVar_p(fpCal->get_var("signalReal_lw")), shape, shape+3, &mapping ));
        mem2disk.push_back( newCdfVectorWriter( &myBench.diag, NcVar_p(fpCal->get_var("signalImag_lw")), shape, shape+3, &mapping ));
        mem2disk.push_back( newCdfVectorWriter( &myBench.resampled, NcVar_p(fpResampled->get_var("signalReal_lw")), shape, shape+3, &mapping ));
        
        // mem2disk.push_back( newCdfCplxVectorWriter( &myBench.zbbSynth, NcVar_p(fpUncal->get_var("zbbReal_lw")), NcVar_p(fpUncal->get_var("zbbImag_lw")), shape, shape+3, &mapping ));
        // mem2disk.push_back( newCdfCplxVectorWriter( &myBench.abbSynth, NcVar_p(fpUncal->get_var("abbReal_lw")), NcVar_p(fpUncal->get_var("abbImag_lw")), shape, shape+3, &mapping ));
        // mem2disk.push_back( newCdfCplxVectorWriter( &myBench.hbbSynth, NcVar_p(fpUncal->get_var("hbbReal_lw")), NcVar_p(fpUncal->get_var("hbbImag_lw")), shape, shape+3, &mapping ));
    }
};


int test( const FileEntrySeq& files )
{
    TestBench myBench;

    // create our science capsules, linking them to test bench buffers
    Ifg2Spc i2sObs(myBench.obsI2sSettings, myBench.obsI2sPorts);
    CalStudy study(myBench.studySettings, myBench.studyPorts);
    CalSynth synth(myBench.synthSettings, myBench.synthPorts);
    CalThreeBody c3b(myBench.calSettings, myBench.calPorts);
    Resampler resample(myBench.resampSettings, myBench.resampPorts);
    
    OutputPack *out = 0;
    
    // iterate through file records; data exports are currently done for each E file.
    for(FileEntrySeq::const_iterator cfi=files.begin(); cfi!=files.end(); ++cfi)
    {
        const FileEntry &entry( *(*cfi) );
                    
        // FUTURE: consider creating alternate mapping which selects a subset of pixels, avoid memory blowout
        GiftsPixelRecordMapping mapping;    
        FileToMemBinderGroup disk2mem;
        long shape[] = { 1, 1, IFG_SIZE };

        disk2mem.push_back( newCdfVectorReader( &myBench.obs, 
                                NcVar_p(entry.nc->get_var("signalReal_lw")), 
                                shape, shape+3, &mapping ));
        disk2mem.push_back( newCdfFieldReader( &myBench.theta,
                                NcVar_p(entry.nc->get_var("Theta")),
                                &mapping ) );
                                
        // if we're on an earth cube, mimic the input file for cal and uncal ouput
        // there's a prettier way to do this, but this'll do
        if (entry.scenetype=='E')
            out = new OutputPack(entry.filename, myBench);
        
        for(unsigned j=0; j<NUM_RECS; j++) // for each pixel in the input file
        {
            disk2mem.update( +1, SEEK_CUR ); // grab next pseudorecord (==pixel) worth of data
            
            // FIXME: check result codes from invocations of algorithms

            switch( entry.scenetype )
            {   
                case 'H':
                case 'A':
                case 'Z':
                    i2sObs();
                    myBench.studyAncil = CalStudy::ancillary_t::fromStruct( entry );
                    // get detector id from the effective record number FUTURE: more elegant way to handle?
                    myBench.studyAncil.detectorid = mapping.detectorid( disk2mem.tell() ); 
                    study();
                    break;
                case 'E':
                    out->mem2disk.update( +1, SEEK_CUR ); // move to next record in disk output binding
                    i2sObs();
                    myBench.synthAncil = CalSynth::ancillary_t::fromStruct( entry );
                    myBench.ancil = CalThreeBody::ancillary_t::fromStruct( entry );
                    myBench.synthAncil.detectorid = mapping.detectorid( disk2mem.tell() );
                    myBench.resampAncil.laserWavenumberRatio = myBench.wnumRatioFromTheta( myBench.theta );
                    synth();
                    c3b();
                    resample();
                    break;
                default:
                    std::cout << "unknown scene type (H/A/Z/E expected\n";
                    abort();
            } //switch
        } // for each pixel
        
        if (out) delete out;
        out = NULL;
        
    // FUTURE: limit the size of (prune) the context cache so as not to overflow memory with superfluous contexts
    // decide whether this function is external or internal to the cache storage itself (probably internal)        
    } // for each data file
    return 0;
}

void printUsageMsg()
{
    std::cout 
        << "Usage: \n" 
        << "   SimCal3Test.unittest [filetype] [filename] [filetype] [filename] ...\n" 
        << "where [filetype] is one of the values:\n"
        << "   H     - hot blackbody data cube\n"
        << "   W     - warm blackbody data cube\n"
        << "   S     - space blackbody data cube\n"
        << "   E     - Earth observation data cube\n"
        << "Options list must specify at least one of each blackbody and at least one observation cube.\n";
}

int main( int argc, char *argv[] )
{
    //physics::wnums_t wnu( TestBench::wavenumbersLW() );
    //cout << "wavenumbers: base " << wnu.baseWnum 
    //        << " interval " << wnu.deltaWnum 
    //        << " count " << wnu.wnumCount << endl;

    // options list must be pairs of options and values
    if(argc%2 != 1) 
    {
        printUsageMsg();
        return 1;
    }
    
    FileEntrySeq files;
    int nH=0,nA=0,nZ=0,nE=0;
    
    for(int i=1; i<argc; i+=2)
    {
        switch( argv[i][0] )
        {
            case 'h':
            case 'H': 
                ++nH;
                files.push_back( FileEntrySeq::value_type( new FileEntry( argv[i+1], 'H' ) ) ); 
                break;
            
            case 'w':
            case 'W': 
                ++nA;
                files.push_back( FileEntrySeq::value_type( new FileEntry( argv[i+1], 'A' ) ) ); 
                break;
            
            case 's':
            case 'S': 
                ++nZ;
                files.push_back( FileEntrySeq::value_type( new FileEntry( argv[i+1], 'Z' ) ) );
                break;

            // FUTURE: o?            
            case 'e':
            case 'E':
                ++nE; 
                files.push_back( FileEntrySeq::value_type( new FileEntry( argv[i+1], 'E' ) ) ); 
                break;
            
            default:
                printUsageMsg();
                std::cout << "ERROR: unknown scene type '" << argv[i] << "', aborting\n";
                abort();
        }            
    }
    
    // encourage that we have matching number of blackbody views
    if ( (nH==0) || (nH!=nA) || (nH!=nZ) || (nE==0) )
    {
            printUsageMsg();
            std::cout << "ERROR: data should have equal number of blackbody views and at least one of each view; aborting\n";
            abort();
    }
            
    return test(files);
}
