/** 
  *  \file FFTResampler.h
  *  \brief Spectral Resampling Algorithm.
  *
  *  \author M. Smuga-Otto <maciek@ssec.wisc.edu>
  *
  *  \version $Id: Resample.hxx,v 1.3.2.3 2006/01/05 00:43:13 maciek Exp $
  *
  *  \par Copyright:
  *  \verbatim
  *
  *  Copyright UW/SSEC, ALL RIGHTS RESERVED, 2004
  *  Space Science and Engineering Center
  *  University of Wisconsin - Madison, USA
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  * 
  * \endverbatim
  *
  * Interface (pretty simple for this stage):
  *
  * obs_in->spectrum[] : spectrally uncalibrated data
  * ref_in->tau[]
  * ref_in->wavenumbers[] : target wavenumbers
  * obs_out->spectrum[]: spectrally calibrated data
  *
  */

#ifndef H_Resample
#define H_Resample

// C++ STL includes
#include <algorithm>
#include <cmath>
#include <functional>
#include <vector>

// GIPS capsule includes
#include "CplxFFT.hxx"

namespace gips {
template <typename InType, typename OutType>
class SpectralResampling
{
public:
    struct settings_t
    {
        unsigned spectrumSize;
        unsigned upsampledSpectrumSize;
    };
    struct ancillary_t
    {
         double laserWavenumberRatio;  
         // note, this may be delivered as some other number or numbers,
         // from which this ratio will need to be computed. I'm going with the simple for now.
     };
     struct Ports
     {
         const ancillary_t * const anc_in;
         const InType * const obs_in;
         OutType * const obs_out;
         Ports(const ancillary_t *anc_in_, const InType *obs_in_, OutType *obs_out_): 
             anc_in(anc_in_), obs_in(obs_in_), obs_out(obs_out_) {}
    };
private:
    typedef std::vector<std::complex<double> > CDVector; 
    typedef CplxFFT<CDVector, CDVector> CDVFFT; 
    
    const settings_t settings;
    const Ports pp;
    const double normalization;
    const int downsampleRatio;
    CDVector mirroredSpectrum, shortIfg, zeroFilledIfg, upsampledSpectrum; // intermediate buffers
    CDVFFT forwardFFT, backwardFFT;  // assume CplxFFT as a subcapsule

public:
    SpectralResampling(const settings_t &s_, const Ports &p_): settings(s_), pp(p_),
         normalization(1.0 / static_cast<double>(settings.spectrumSize * 2)),
         downsampleRatio(settings.upsampledSpectrumSize / settings.spectrumSize / 2),
         mirroredSpectrum(settings.spectrumSize * 2),
         shortIfg(settings.spectrumSize * 2), 
         zeroFilledIfg(settings.upsampledSpectrumSize, 0.0), // ensure this is zeroed out
         upsampledSpectrum(settings.upsampledSpectrumSize), // surpising, that one ;-)
         forwardFFT(
                 CDVFFT::settings_t(settings.spectrumSize * 2, 
                                                 CDVFFT::FORWARD,  // define this enum member in CplxFFT
                                                 CDVFFT::ESTIMATE),   // ditto
                 CDVFFT::Ports(&mirroredSpectrum, &shortIfg)),
         backwardFFT(
                 CDVFFT::settings_t(settings.upsampledSpectrumSize,
                                                 CDVFFT::BACKWARD,
                                                 CDVFFT::MEASURE),
                 CDVFFT::Ports(&zeroFilledIfg, &upsampledSpectrum))
    {} 
    ~SpectralResampling() {}
    void operator()(void) 
    {
        const InType &obsIn (*(pp.obs_in)); // avoid convoluted pointer deference syntax in subsequent client code.

        std::copy(obsIn.begin(), obsIn.end(), mirroredSpectrum.begin());
        std::copy(obsIn.begin() + 1, obsIn.end(), mirroredSpectrum.rbegin());

        forwardFFT(); // reap benefits of CplxFFT capsule

        const CDVector::iterator shortIfgMiddle(shortIfg.begin() + settings.spectrumSize);
        const CDVector::reverse_iterator shortIfgRMiddle(shortIfg.rbegin() + settings.spectrumSize - 2);

        std::copy(shortIfg.begin(), shortIfgMiddle, zeroFilledIfg.begin());
        std::copy(shortIfg.rbegin(), shortIfgRMiddle, zeroFilledIfg.rbegin());
        zeroFilledIfg[settings.spectrumSize -1] *= 0.5; // weird, but I had it in the original code...
        backwardFFT();

        const CDVector::iterator upsampledSpectrumMiddle(upsampledSpectrum.begin() 
                                     + settings.upsampledSpectrumSize / 2);
        std::transform(upsampledSpectrum.begin(), upsampledSpectrumMiddle, upsampledSpectrum.begin(),
                          std::bind2nd(std::multiplies<std::complex<double> >(), normalization));
  
        // the following code is directly transcribed from the old style FFTResampler.cc source
        double a = downsampleRatio * (settings.spectrumSize - 1) *  ( 1 / pp.anc_in->laserWavenumberRatio - 1 );
        double b = downsampleRatio / pp.anc_in->laserWavenumberRatio;
        double r, rf; // fractional resample position
        int ri; // integral resample position

        OutType &obsOut (*(pp.obs_out));

        for (unsigned int n = 0; n < settings.spectrumSize; ++n)
        {
            r = a + b * n;
            ri = static_cast<int>(floor(r));
            rf = r - ri;

            obsOut[n] = (r < 0 || r >= settings.upsampledSpectrumSize / 2)
                              ? 0
                              : std::real(upsampledSpectrum[ri]) 
                                + rf * (std::real(upsampledSpectrum[ri+1]) 
                                - std::real(upsampledSpectrum[ri]));
        }
    } // end operator()
}; // end class SpectralResampling
} // end namespace gips
#endif
